from typing import Callable, Any
from pathlib import PurePosixPath 
from xingyun.cloud.egorov_awss3 import gets3, sets3

class DataAccess:
    '''This class defines how to access data that is stored remotely.'''

    def __init__(self,
        path: PurePosixPath | str, 
        get_call: Callable[ [str], Any ] , 
        set_call: Callable[ [str, Any], bool ] , 
    ):
        '''The intialization function of class `DataAccess`.

        ### Parameters
            - path: The name independent path to quries. The actual path will be `path / key`.  
                If `None`, will defaultly use the data saved in the `xingyun-GlobalDataManager` of EgorovSystem.
            - get_call: A callable object. The method to get data remotely. 
                The first parameter of `get_call` should be the **absolute** path of the data.
            - set_call: A callable object. The method to set data remotely. 
                The first parameter of `set_call` should be the **absolute** path of the data. 
                The second parameter of `set_call` should be the value to set. 
                Returns a bool to show if the set is successful.
        '''
        if isinstance(path , str):
            path = PurePosixPath(path)
        self.path = path
        self.get_call = get_call
        self.set_call = set_call

    def cd(self, name: str):
        '''This function returns a copy of `self`, with path updated to `self.path / name`. 
        '''
        return DataAccess(self.path / name, self.get_call, self.set_call)
    
    def set_path(self, new_path: str | PurePosixPath | None):
        '''This function returns a copy of `self`, with path updated to `new_path`. 
        '''
        if new_path is None:
            return None
        if isinstance(new_path , str):
            new_path = PurePosixPath(new_path)
        return DataAccess(new_path, self.get_call, self.set_call)

    def get(self, key: str):
        return self.get_call(str(self.path / key))
    
    def set(self, key: str, val: Any):
        return self.set_call(str(self.path / key), val)

def get_aws3_dataacess():
    return DataAccess(
        PurePosixPath("") , 
        lambda k: gets3(k, format = "pickle"), 
        lambda k,v: sets3(v,k, format = "pickle")
    )



'''This module optionally denpends on the `xingyun-getid` entry of [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).
### Egorov System Data Format
    `{path}`
'''
from egorovsystem import get_variable
from typing import Callable, Any
from .base import get_aws3_dataacess, DataAccess

_my_default_dataacess = get_aws3_dataacess().set_path(get_variable("xingyun-getid"))

def get_id(
    name: str, 
    data_access: DataAccess | None = None, 
) -> str:
    '''This function ensures every call with the same `path / name` gets a fresh number.
    
    This function optionally denpends on the `xingyun-getid` entry of 
    [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).

    ### Egorov System Data Format
        `{path}`

    ### Parameters
        - name: project name.
        - data_access: A `DataAccess` instance to guide how to save data.
    '''

    if data_access is None:
        if _my_default_dataacess is None:
            raise RuntimeError("no meta_path. no egorov system entry.")
        data_access = _my_default_dataacess

    val = data_access.get(name)
    if val is None:
        val = -1
    val = val + 1
    if not data_access.set(name, val):
        raise RuntimeError("set failed.")
    return str(val)
    
    


from pathlib import Path
import re
from typing import Callable
import fnmatch
from pathlib import PurePosixPath
from xingyun.universal.convert import convert_size_to_bytes

def filter_default(filename: str):
    '''filter file based on file name.'''

    return re.search(r"((\.py)|(\.sh)|(\.txt))$", filename) is not None


def filter_hidden(filename: str):
    '''skip files that are start with "." .'''

    return re.search(r"(/|^)\.[^\\]", filename) is None

def filter_gitignore(filename: str):
    '''filter out files that are in .gitignore'''

    p = Path(".gitignore")
    if p.exists():
        with open(p, "r") as fil:
            ignore_files = fil.read().split("\n")
        for to_ignore in ignore_files:
            if fnmatch.fnmatch(filename, to_ignore):
                return False
    return True

def get_code(
        filters: list[ Callable[[str], bool] ] = [filter_default, filter_gitignore, filter_hidden], 
        sizelimit : int | str = "1mb" , 
        total_sizelimit  : int | str = "500mb" , 
        path: str = ".", 
    ) -> dict[str, str]:
    '''Get all code files under a dictionary.
    
    ### Parameters
        - path: path to start traverse.
        - pattern: a regular expression to match files.
        - sizelimit: only save files that are lower than size limit.
    '''
    p = Path(path)
    if isinstance(sizelimit , str):
        sizelimit = convert_size_to_bytes(sizelimit)
    if isinstance(total_sizelimit , str):
        total_sizelimit = convert_size_to_bytes(total_sizelimit)

    saved = {}
    acc_size = 0 # accumulated size
    for file in p.rglob("*"):

        # ensure file
        if not file.is_file():
            continue 

        # get relative path
        rel_path = str( PurePosixPath(file.relative_to(p)) )

        if rel_path == "__init__.py":
            import pdb;pdb.set_trace()

        # apply filters
        flag_skip = False
        for filter in filters: # skip if can not pass all filters
            if not filter(rel_path):
                flag_skip = True
        if flag_skip:
            continue


        # apply filesize limit
        file_size = file.stat().st_size
        if file_size > sizelimit: # skip if too large 
            continue

        acc_size = acc_size + file_size
        if acc_size > total_sizelimit: # break if total too large
            break

        # save file content
        with open(file, "rb") as fil:
            content = fil.read()

        try:
            content = content.decode("utf-8")
        except:
            pass
        
        saved[rel_path] = content

    return saved

def compare_dict(dict_1: dict[str, str] | None, dict_2: dict[str, str] | None):
    '''compare if two str dicts are exactly the same.'''

    if (dict_1 is None) or (dict_2 is None):
        return (dict_1 is None) and (dict_2 is None)
    
    names = list( set(dict_1) | set(dict_2) )
    names.sort()
    hash_1 = "##".join([str( dict_1.get(x) ) for x in names])
    hash_2 = "##".join([str( dict_2.get(x) ) for x in names])
    return hash_1 == hash_2



'''This module provides methods to manage and save data.'''

from .base import DataAccess, get_aws3_dataacess
from .getid import get_id
from .globaldata import GlobalDataManager, make_hook_logger, HookType
from .logger import Logger



import time

_starttime = time.time()

def my_clock():
	'''call time.time() to see current running time. 
	This is more accurate than time.clock().
	
	XXX really?
	'''
	return time.time() - _starttime



from typing import Any
 
def get_subdict(d: dict[str, Any], prefix: str, splitter: str = "/"):
    '''Find a subdict from a dict.

    ### Example
        >>> dic = {"a": 1, "a/b": 2, "b": 3}
        >>> get_subdict(dic, "a")
        {"b": 2}
    
    '''
    ret = {}
    for x,y in d.items():
        splitted = x.split(splitter)
        if splitted[0] == prefix:
            ret[splitter.join(splitted[1:])] = y
    return ret


_seed = 2333
def my_rand():
    global _seed
    _seed = _seed * 233 + 23333
    _seed = _seed % 233333333
    return _seed

def my_randint(low: int, high: int):
    if high - low <= 0:
        return low
    r = my_rand()
    return (r % (high - low)) + low


'''This module optionally denpends on the `xingyun-GlobalDataManager` entry of [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).
### Egorov System Data Format
    `{path}`
'''
from egorovsystem import get_variable
from typing import Any, Union, Callable, List, TypeAlias
import atexit
from xingyun.savecode.get_code import get_code, filter_default, filter_gitignore, compare_dict, filter_hidden
from .logger import Logger
from .base import get_aws3_dataacess, DataAccess

HookType: TypeAlias = Callable[[str,Any,int,dict,str], Any]

def call_hook(hk: HookType, key: str, val: Any, timestamp: int, all_data: dict, modespace: str):
    '''call a hook.
    This function is just to make my auto complete system work when coding.'''
    return hk(key, val, timestamp, all_data, modespace)

def make_hook_logger(logger: Logger):
    '''This is a hook helper funtion. It helps to create a hook that can be used for `GlobalDataManager`.
    '''
    def _logger_hook(key: str, val: str, timestamp: int, all_data: dict, modespace: str):
        if modespace == GlobalDataManager.LOGGER_MODESPACE:
            logger.log(val)
    return _logger_hook

class GlobalDataManager:
    '''This class is used to save all important data of a process. It will automatically upload data to S3. 
        This class will create a meta file under given path to store meta information (list of keys).

    This class optionally denpends on the `xingyun-GlobalDataManager` entry of 
    [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).
    
    ### Egorov System Data Format
        `{path}`
    '''
    METAFILE = "_meta"
    DEFAULT_MODESPACE = "default"
    LOGGER_MODESPACE = "logger"
    LOGGER_KEY = "_log"
    CODE_KEY = "_code"
    EGO_KEY = "xingyun-GlobalDataManager"

    def __init__(self, 
        name: str , 
        hooks: List[HookType] = [] , 
        data_access: DataAccess | None = None , 
        sync_time: int = 50
    ):
        '''
        The intialize function of vlass `GlobalDataManager`.

        ### Parameters
            - name_: the name of the current project.
            - hooks: a list of hooks. Each hook is a callable object that will be called at each time when data updates. 
                The parameters of a hook call is `(key, value, timestamp, all_data, modespace)`. 
                The parameter `modespace` is used to control which hooks to be called.
            - data_access: A `DataAccess` instance to guide how to save data.
            - sync_time: will automatically sync every `sync_time` times. 
        '''

        # ensure data access
        _my_default_dataacess = get_aws3_dataacess().set_path(get_variable(type(self).EGO_KEY))
        if data_access is None:
            if _my_default_dataacess is None:
                raise RuntimeError("no path. no data in EgorovSystem.")
            data_access = _my_default_dataacess
        self.data_access = data_access.cd(name)

        # initialize / recover data
        self.datas: dict[str, dict[int, Any]] = {}
        self.time_stamps: dict[str, int] = {}
        if not self.download(): # if no remote data, then create a meta entry
            self.update_meta()

        # store hooks
        self.hooks = hooks

        # intialize sync params
        self.sync_time = max(int(sync_time), 1) # ensure no devide by 0
        self._sync_count = 0
            
    def __del__(self):
        # ensure sync at exit
        self.upload_data()

    def should_sync(self, sync: bool | None):
        ''' Return true one time every `self.sync_time` calls.
        ### Parameters
            - sync: If not `True`, then sync whatever. If not `False`, then don't sync whatever. 
        '''
        self._sync_count = (self._sync_count + 1) % self.sync_time
        if sync is None:
            return self._sync_count == 0
        return sync

    def upload_data(self, keys: list[str] | None = None) -> bool:
        '''Upload data to remote.
        
        ### Parameters
            - keys: the keys of the data to upload. If `None`, then all data.
        '''
        flag = True
        flag = flag and self.update_meta()

        if keys is None:
            keys = list(self.datas)
        for key in keys:
            flag = flag and self.set_remote(key, self.datas.get(key))
        return flag
    
    def update_meta(self) -> bool:
        '''If the meta file is not created, then creat it.'''

        return self.set_remote(GlobalDataManager.METAFILE, {
            "time_stamps": self.time_stamps , 
        })
    
    def download(self):
        '''Read the meta file information and recover data. If no meta file then return `False`.'''

        meta_info = self.get_remote(GlobalDataManager.METAFILE)
        if meta_info is None:
            return False
        
        try:
            time_stamps: dict[str, int] = meta_info["time_stamps"]
        except:
            return False
        
        self.time_stamps = time_stamps
        for key in time_stamps:
            self.datas[key] = self.get_remote(key)


    def update_when_diff(self, 
        key: str, 
        val: Any,
        cmp: Callable[[Any, Any], bool] , 
        modespace: str = DEFAULT_MODESPACE , 
        overwrite: bool = False, 
        force_timestamp: int | None = None, 
        sync: bool | None = None
    ) -> bool:
        '''Set data. But only sets when data is different from stored one.

        ### Parameters
            - force_timestamp: If not `None`, then will force timestamp setting to this value.
            - overwrite: If `True`, will not update timestamp. Else timestamp will be automatically shifted by `1`. 
                If `force_timestamp == True`, then this parameter will not be used.
            - sync: If `True`, then will upload the data to remote. If `None` then use the `autosync` property.
        '''

        time_stamp = self.time_stamps.get(key)
        if force_timestamp is not None: # force time stamp
            time_stamp = force_timestamp

        saved = self.get(key, time_stamp)
        if cmp(saved, val):
            return True
        return self.set(key, val, modespace, overwrite, force_timestamp, sync)

    def set(self, 
        key: str, 
        val: Any, 
        modespace: str = DEFAULT_MODESPACE , 
        overwrite: bool = False, 
        force_timestamp: int | None = None, 
        sync: bool | None = None
    ) -> bool:
        '''Set data. 

        ### Parameters
            - overwrite: If `True`, will not update timestamp. Else timestamp will be automatically shifted by `1`. 
                If `force_timestamp == True`, then this parameter will not be used.
            - force_timestamp: If not `None`, then will force timestamp setting to this value.
            - sync: If `True`, then will upload the data to remote. If `None` then use the `autosync` property.
        '''

        # ensure and update timestamp
        time_stamp = self.time_stamps.get(key)
        if force_timestamp is not None:         # force time stamp anyway
            time_stamp = force_timestamp
        else:
            if time_stamp is None:              # set default time stamp
                time_stamp = 0
            else:
                if not overwrite:               # apply only when time_stamp is already created       
                    time_stamp = time_stamp + 1 # increment time stamp

        self.time_stamps[key] = time_stamp

        # ensure data entry
        if self.datas.get(key) is None:
            self.datas[key] = {}
        data_entry = self.datas.get(key)

        if not isinstance(data_entry, dict):
            raise RuntimeError(f"THIS CAN NEVER HAPPEN.")

        # update data
        data_entry[time_stamp] = val

        # optionally sync data
        flag = True
        if self.should_sync(sync):
            flag = flag and self.upload_data([key])

        # call hooks
        for hk in self.hooks:
            call_hook(hk, key, val, time_stamp, data_entry, modespace)

        return flag
    
    def get(self, key: str, time_stamp: int | None = None):
        '''Get data. If no data return `None`.

        ### Parameters
            - time_stamp: Ask the data at a specific time stamp. If `None`, then default to current time stamp.
        '''
        
        if time_stamp is None:
            time_stamp = self.time_stamps.get(key)

        if time_stamp is None:
            return None
        
        data_entry = self.datas.get(key)
        if data_entry is None:
            return None
        
        return data_entry.get(time_stamp)

    def get_remote(self, key: str):
        return self.data_access.get(key)
    
    def set_remote(self, key: str, val: Any):
        return self.data_access.set(key, val)


    def log(self, content: str, sync: bool | None = None):
        '''This function save data to logger modespace.'''
        return self.set(type(self).LOGGER_KEY, content, modespace = type(self).LOGGER_MODESPACE, sync = sync)

    def get_timestamp(self, key: str) -> int:
        '''Returns the current timestamp of a file. If it is not created, return -1.'''
        t = self.time_stamps.get(key)
        if t is None:
            return -1        
        return t

    def exists(self, key: str):
        return self.datas.get(key) is None

    def summarize(self, key: str, summarize_func: Callable[[dict[int, Any]] , Any]):
        '''Summerize all of the history of a data entry.'''
        data_entry = self.datas.get(key)
        if data_entry is None:
            return None
        return summarize_func(data_entry)
        
    def log_code(self, 
        filters: list[ Callable[[str], bool] ] = [filter_default, filter_gitignore, filter_hidden], 
        sizelimit : int | str = "1mb" , 
        total_sizelimit : int | str = "500mb" , 
        path: str = ".", 
    ):
        '''This function log current code. Only update when there is change in code. Returns code version.'''
        code_data = get_code(filters, sizelimit, total_sizelimit, path)
        self.update_when_diff(type(self).CODE_KEY, code_data, compare_dict, "code")
        return self.get_timestamp(type(self).CODE_KEY)



import time
from timeit import default_timer as gettime
from prettytable import PrettyTable

class Timer:
	'''一个计时器。

	用法：
	with Timer(name):
		do_someting

	do_someting这段代码的用时会以name为键记录。
	Timer.output_all()会生成一个字符串描述已有的所有记录。

	示例：
	with Timer("hahah"):
		print (123)
	with Timer("sleep"):
		time.sleep(1)
	print(Timer.output_all())
	'''

	time_sum = {}
	count = {}
	before_enter = []
	before_exit = []

	def __init__(self , name):
		self.name = name

	def __enter__(self):
		for x in Timer.before_enter:
			x()
		self.start_time = gettime()
		return self

	def __exit__(self , *args , **kwargs):
		for x in Timer.before_exit:
			x()
		the_time = gettime() - self.start_time

		if self.time_sum.get(self.name) is None:
			self.time_sum[self.name] = 0
			self.count[self.name] = 0
		
		self.time_sum[self.name] += the_time
		self.count[self.name] += 1

	@classmethod
	def get_time(cls, name: str , average = False) -> float:
		if Timer.time_sum.get(name) is None:
			return -1
		ret = Timer.time_sum[name]
		if average:
			ret /= Timer.count[name]
		return ret

	@classmethod
	def get_avg_time(cls, name: str) -> float:
		return Timer.get_time(name , True)
	
	@classmethod
	def clear(cls):
		Timer.time_sum = {}
		Timer.count = {}

	@classmethod
	def get_all(cls, average = False) -> list[float]:
		time_sum = Timer.time_sum
		count = Timer.count

		ret = []
		for x in time_sum:
			if average:
				ret.append([x , time_sum[x] , count[x]])
			else:
				ret.append([x , time_sum[x] / count[x] , count[x]])

		return ret

	@classmethod
	def output_all(cls):
		def f2s(f):
			return "%.10f" % f
		time_sum = Timer.time_sum
		count = Timer.count

		table = PrettyTable(["name" , "time (s)" , "count" , "avg_time (s)"])
		for x in time_sum:
			table.add_row([x , f2s(time_sum[x]) , count[x] , f2s(time_sum[x] / count[x])])

		return str(table)




from importlib import import_module as impo
from types import ModuleType
import warnings 


def my_import_module(module_name: str) -> ModuleType | None:
    try:
        lib = impo(module_name)
    except:
        warnings.warn(f"module {module_name} does not exists.")
        return None

    return lib


'''
This module make using of AWS S3 even easier. It reads the aws access id and 
key from EgorovSystem.

This module denpends on the `aws_access` entry of [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).
### Egorov System Data Format
    "{key_id} {key_sec} {region} {bucket}"
'''

from egorovsystem import Egorov, get_variable
import boto3
from typing import Literal, Any
from .awss3 import AWSS3

_awss3_initialized = False
_awss3 = None

def get_awss3_instance():
    global _awss3
    global _awss3_initialized
    
    if not _awss3_initialized:
        try: 
            key_id , key_sec , region , bucket = get_variable("aws_access").split(" ")
            _awss3 = AWSS3(key_id, key_sec, region, bucket)
        except Exception:
            _awss3 = None
    
    return _awss3

def sets3(data: Any, tar_path: str, format: Literal["binary" , "str" , "pickle"] = "pickle"):
    '''This function denpends on the `aws_access` entry of [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).
    ### Egorov System Data Format
        `{key_id} {key_sec} {region} {bucket}`
    '''

    awss3 = get_awss3_instance()
    if awss3 is None:
        raise RuntimeError("no aws account found.")
    
    return awss3.set(data, tar_path, format)

def gets3(tar_path: str, format: Literal["binary" , "str" , "pickle"] = "pickle") -> Any:
    '''This function denpends on the `aws_access` entry of [EgorovSystem](https://github.com/FFTYYY/EgorovSystem).
    ### Egorov System Data Format
        `{key_id} {key_sec} {region} {bucket}`
    '''

    awss3 = get_awss3_instance()
    if awss3 is None:
        raise RuntimeError("no aws account found.")
    
    return awss3.get(tar_path, format)



import sys
from typing import Callable, Any
import re
import warnings
from argparse import Namespace
from ..random.my_random import my_randint
from ..universal.get_subdict import get_subdict

class Condition:
    def __init__(self, condition: Callable[[ dict[str, Any] ], bool] = lambda _: True):
        self.condition = condition

    def test(self, config: dict[str, Any]):
        return self.condition(config)
    
    def __and__(self, other: "Condition"):
        return Condition(lambda v: self.condition(v) and other.condition(v))
    
    def __or__(self, other: "Condition"):
        return Condition(lambda v: self.condition(v) or other.condition(v))
    
    @classmethod
    def and_all(cls, conditions: list["Condition"]):
        r = Condition()
        for c in conditions:
            r = r & c
        return r


condition_env = {}
class PreCondition:
    def __init__(self, cond: Condition | Callable[[ dict[str, Any] ], bool]):

        if not isinstance(cond, Condition):
            cond = Condition(cond)

        self.cond = cond
        self.id = my_randint(0,2333333)
    def __enter__(self):
        condition_env[self.id] = self.cond
    def __exit__(self, *arg, **kwarg):
        condition_env.pop(self.id)

class MyDict(dict):
    def __init__(self, d: dict = {}):
        super().__init__(d)

    def sub(self, prefix: str, splitter: str = "/"):
        return MyDict( get_subdict(self, prefix, splitter) )
    
    def __call__(self, prefix: str, splitter: str = "/"):
        return self.sub(prefix, splitter)
    

class Argument:
    '''The class that describe an argument.
    
    ### Properties
        - name: the name of the argument.
        - type: how to convert string to a value.
        - default: the value set to the argument if it is not provided.
        - present: the value set to the argument if it is provided but without a value.
        - help: help string.
        - pre_condition: Before an argument is parsed, the pre-condition must be satisfied.
        - post_condition: After an argument is parsed, if the post-condition is not satisfied, then will raise an error. 
    '''
    def __init__(self, 
            name: str, 
            type: Callable[[str] , Any], 
            default: Any = None, 
            present: Any = None,
            help:str = "" , 
            pre_condition : Condition = Condition(), 
            post_condition: Condition = Condition(), 
        ):
        self.name = name
        self.type = type
        self.default = default
        self.present = present
        self.pre_condition  = pre_condition
        self.post_condition = post_condition
        self.help = help

class ArgumentParser:
    '''This class is a modification of python `argparse.ArgumentParser` class.

    For each argument, there are two conditions: pre-condition and post-condition.
    Before an argument is parsed, the pre-condition must be satisfied. 
    After an argument is parsed, if the post-condition is not satisfied, then will raise an error.
    '''

    def __init__(self, help: str = ""):

        self.help = help

        self.arguments: dict[str, Argument] = {}
        self.alias    : dict[str, str] = {} # redirects to

    @property
    def now_preconds(self):
        return [c for id,c in condition_env.items() if c is not None]
    
    @classmethod
    def get_subconfig(cls, C: dict | Namespace, prefix: str, splitter: str = "/"):
        if isinstance(C, Namespace):
            C = C.__dict__
        return get_subdict(C,prefix,splitter)

    def add_alias(self, alias: str, original: str):
        '''add an alias for an argument.'''
        self.alias[alias] = original
        
    def add_argument(self, 
        name: str, 
        type: Callable[[str] , Any], 
        default: Any = None, 
        present: Any = None, 
        help: str = "" , 
        verify : Callable[ [Any] , bool] = lambda _: True ,
        aliases: list[str] = []
    ):
        post_cond = Condition(lambda C: verify(C.get(name)))
        self.arguments[name] = Argument(name, type, default, present, help, Condition.and_all(self.now_preconds), post_cond)
        for alias in aliases:
            self.add_alias(alias, name)

    def add_bool(self, 
        name: str, 
        help: str = "" , 
        verify : Callable[[ Any ], bool] = lambda _: True ,
    ):
        self.add_argument(name, bool, False, True, help, verify)

    def parse_namespace(self, 
            args: list[str] | None = None, 
            pattern = r"^--([^=]+)(=(.+)|)$", 
            get_match: Callable[[re.Match], tuple[str,str]] = lambda m: (m.group(1), m.group(3)) , 
        ) -> Namespace:
            return Namespace(**self.parse(args, pattern, get_match))
    
    def parse(self, 
            args: list[str] | None = None, 
            pattern = r"^--([^=]+)(=(.+)|)$", 
            get_match: Callable[[re.Match], tuple[str,str]] = lambda m: (m.group(1), m.group(3)) , 
        ) -> MyDict:
        ''' Parse argument.

        ### Parameters
            -- args: arguments to be parsed.
            -- pattern: a regular expression to match name and value.
            -- get_match: a callable object that get name and value from the get_match. The input callable is a `re.Match` object, 
                the output should be a 2-tuple, with the first element being the name and the second element being the value of the argument.
        '''
        if args is None:
            args = sys.argv[1:]

        # get value of each argument 
        name_vals = {}            
        for s in args:
            # get name val pairs
            match = re.match(pattern, s)
            if match is None:
                continue
            name, val = get_match(match)

            # apply alias 
            alias_tar = self.alias.get(name)
            if alias_tar is not None:
                name = alias_tar
            
            # apply present val
            arg = self.arguments.get(name)
            if arg is None:
                continue
            if val is None:
                val = arg.present

            # record name, val pair
            name_vals[name] = val
        
        # apply default val
        for name , arg in self.arguments.items():
            if not (name in name_vals):
                name_vals[name] = arg.default

        # actually assign values 
        parsed = {}
        _t = 0
        for _t in range(100):
            now_len = len(parsed)
            for name, val in name_vals.items():

                # get Argument object
                arg = self.arguments.get(name)
                if arg is None:
                    continue
                     
                # check pre condition
                if not arg.pre_condition.test(parsed):
                    continue
                
                # store value
                parsed[name] = arg.type(val)

                # check post condition
                if not arg.post_condition.test(parsed):
                    raise RuntimeError(f"bad argument {name}: value can not be {parsed[name]}")

            new_len = len(parsed)
            if new_len == now_len: # no new argument is parsed
                break
            now_len = new_len

        if _t >= 98:
            warnings.warn("Too deep nested logic. Only performed 100 iterations.")
        
        # for those who forbidded by precondition, assign `None`.
        for name in self.arguments:
            if not (name in parsed):
                parsed[name] = None 

        return MyDict(parsed)





from setuptools import setup, find_packages

with open("README.md", encoding="utf-8") as f:
    readme = f.read()

with open("LICENSE", encoding="utf-8") as f:
    license = f.read()

reqs = ""
try:
    with open("requirements.txt", encoding="utf-8") as f:
        reqs = f.read()
except:
    print("fuck")

pkgs = [p for p in find_packages() if p.startswith("xingyun")]

setup(
    name="xingyun",
    version="0.0.5",
    url="http://github.com/FFTYYY/XingYun",
    description="",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT",
    author = "Yongyi Yang",
	author_email = "yongyi@umich.edu",
    python_requires = ">=3.10",
    packages = pkgs,
    install_requires = reqs.strip().split("\n"),

    include_package_data = True , 
)